/*  <lalVerbatim file="LALNoiseSpectralDensityCV">
Author: Sathyaprakash, B. S.
$Id: LALNoiseSpectralDensity.c,v 1.6 2002/01/28 12:29:11 dchurches Exp $
</lalVerbatim>  */
/* <lalLaTeX>
\subsection{Module \texttt{LALNoiseSpectralDensity.c}}
This module generates an array of size specified
in the vector \texttt{psd}, that is \texttt{psd.length}.
The inputs are 
\begin{enumerate}
\item The lenght of the psd array required: this must be
given as a non-zero positive integer by setting the \texttt{length}
of the \texttt{psd} vector to the desired value;
\item Frequency resolution \texttt{df} in Hz.
\item Pointer to a function that should be used in generating the
power spectral density values in units of Hz$^{-1}.$ This function
must necessarily be of the type:
\texttt{ void  (*NoisePsd)(LALStatus *status, REAL8 *shf, REAL8 f).}
\end{enumerate}
Presently, there are four such functions in the \texttt{noisemodels}
package. These are \texttt{LALGEOPsd, LALLIGOIPsd, LALTAMAPsd, LALVIRGOPsd.}
These four packages return a scaled PSD while this module returns the
correctly scaled version. It is assumed that new PSD modules return
unscaled versions. (Note, however, that it might be better to use the
scaled versions of the PSD when computing the metric on the signal 
manifold; this is because computing the metric involves calculation of
many moments of the noise PSD and one might encounter round-off errors
if un-scaled version of PSD is used; I have not checked this to be
the case but suspect that there might be some problems.)

\subsubsection*{Prototypes}
\vspace{0.1in}
\input{LALNoiseSpectralDensityCP}
\idx{LALNoiseSpectralDensity()}

\subsubsection*{Description}
\subsubsection*{Algorithm}
\subsubsection*{Uses}
Uses a user specified pointer to a function of type
\begin{verbatim}
void  (*NoisePsd)(LALStatus *status, REAL8 *shf, REAL8 f),
\end{verbatim}
that returns PSD values in units of Hz$^{-1}.$

\subsubsection*{Notes}

\vfill{\footnotesize\input{LALNoiseSpectralDensityCV}}
</lalLaTeX>  */
#include <lal/LALNoiseModels.h>

NRCSID (LALNOISESPECTRALDENSITYC, "$Id: LALNoiseSpectralDensity.c,v 1.6 2002/01/28 12:29:11 dchurches Exp $");

/*  <lalVerbatim file="LALNoiseSpectralDensityCP"> */
void 
LALNoiseSpectralDensity 
   (
   LALStatus    *status, 
   REAL8Vector  *psd, 
   void         (*NoisePsd)(LALStatus *status, REAL8 *shf, REAL8 f),
   REAL8        df
   ) 
{  /*  </lalVerbatim>  */

    REAL8 f, shf, fs, s0;
    INT4 i, n;

   INITSTATUS(status, "LALNoiseSpectralDensity", LALNOISESPECTRALDENSITYC);
   ATTATCHSTATUSPTR(status);

   ASSERT (psd->data,  status, LALNOISEMODELSH_ENULL, LALNOISEMODELSH_MSGENULL);
   ASSERT (NoisePsd, status, LALNOISEMODELSH_ENULL, LALNOISEMODELSH_MSGENULL);
   ASSERT (df > 0., status, LALNOISEMODELSH_ESIZE, LALNOISEMODELSH_MSGESIZE);
   ASSERT (psd->length > 1, status, LALNOISEMODELSH_ESIZE, LALNOISEMODELSH_MSGESIZE);

   /*
   switch (NoisePsd)
   {
	case LALGEOPsd:
           s0 = 1.e-46;
           fs = 10.;
	   break;
        case LALLIGOIPsd:
           s0 = 9.0e-46;
           fs = 10.;
	   break;
        case LALVIRGOPsd:
           s0 = 3.24e-46;
           fs = 10.;
	   break;
        case LALTAMAPsd:
           s0 = 75.e-46;
           fs = 75.;
	   break;
	default:
           s0 = 1.;
           fs = 0.;
	   break;
   }
   */
   if (NoisePsd == LALGEOPsd) {
           s0 = 1.e-46;
           fs = 10.;
   } else if(NoisePsd == LALLIGOIPsd) {
           s0 = 9.0e-46;
           fs = 10.;
   } else if(NoisePsd == LALTAMAPsd) {
           s0 = 75.e-46;
           fs = 75.;
   } else if(NoisePsd == LALVIRGOPsd) {
           s0 = 3.24e-46;
           fs = 10.;
   } else {
           s0 = 1.;
           fs = 0.;
   }

   n = (INT4) psd->length-1;

   /* 
    * Set DC and Nyquist components to zero
    */

   psd->data[0] = psd->data[n] = 0.;

   for (i=1; i<n; i++) 
   {
      f = i*df;

      if (f>fs) 
      {
         (*NoisePsd)(status->statusPtr, &shf, f);
         psd->data[i] = s0 * shf;

      } else {

         psd->data[i] = 0.;
      } 
   }
   DETATCHSTATUSPTR(status);
   RETURN(status);
}

